package MIDICompose;

import java.util.*;

public class MIDIContainer extends MIDIObject implements Channelable, Lengthable
{
	public static final int FREE = 1;			// all elements can be turnedRound and be lengthen
	public static final int LINKED = 2;		// all elements can be lengthen but not turned
	public static final int ABSOLUTE = 3;	// no element can be lengthen or turnedRound

	private boolean linked;								// true: turnRound is impossible
	private boolean lengthable;						// true: elements can be lengthen
	private boolean channelable;					// true: channel of all elements is set
	private int channel;									//       to this channel

	public MIDIContainer() { this(0, FREE); }
	public MIDIContainer(int channel) { this(0, FREE, channel); }
	public MIDIContainer(int deltaTime, int type, int channel) {
		this(deltaTime, type); setChannelable(channel); }
	public MIDIContainer(int deltaTime, int type) {
		super(deltaTime);
		linked = false;
		setType(type);
		channelable = false; }

	protected MIDIContainer(MIDIContainer object) {		// returns a copy of object
	  this(object.deltaTime(), object.type());
		if(object.channelable()) setChannelable(object.channel());
		for(int i = 0; i < object.size(); i++) add((object.elementAt(i)).copy());
		if(object.locked()) lock(); }


//******************************* Channelable routines ***********************************


	public boolean channelable() { return channelable; }
	public void setChannelable(int channel) { channelable = true; setChannel(channel); }

	public int channel() { if(channelable) return channel; else return -1; }
	public void setChannel(int channel) {
		if(!channelable) return;									// no effect if not channelable
		for(int i = 0; i < size(); i++) setChannelAt(channel, i);
		this.channel = channel; }

	public int channelAt(int index) {									// returns channel of element at index
		if(elementAt(index) instanceof Channelable)
			return ((Channelable)elementAt(index)).channel();
		else return -1; }
	public void setChannelAt(int channel, int index) {// sets channel of element at index
		if(elementAt(index) instanceof Channelable)
			((Channelable)elementAt(index)).setChannel(channel); }


//******************************** Amountable routines ***********************************


	public int amountAt(int index) {									// returns amount of element at index
		if(elementAt(index) instanceof Amountable)
			return ((Amountable)elementAt(index)).amount();
		else return -1; }
	public void setAmountAt(int amount, int index) {	// sets amount of element at index
		if(elementAt(index) instanceof Amountable)
			((Amountable)elementAt(index)).setAmount(amount); }


//********************************* Pitchable routines ***********************************


	public int pitchAt(int index) {										// returns pitch of element at index
		if(elementAt(index) instanceof Pitchable)
			return ((Pitchable)elementAt(index)).pitch();
		else return -1; }
	public void setPitchAt(int pitch, int index) {		// sets pitch of element at index
		if(elementAt(index) instanceof Pitchable)
			((Pitchable)elementAt(index)).setPitch(pitch); }


//******************************** Lengthable routines ***********************************


	public int lengthAt(int index) {									// retuns length of element at index
		if(elementAt(index) instanceof Lengthable)
			return ((Lengthable)elementAt(index)).length();
		else return 0; }
	public void setLengthAt(int length, int index) {	// sets length of element at index
		if(elementAt(index) instanceof Lengthable)
			((Lengthable)elementAt(index)).setLength(length); }

	public int deltaTimeAt(int index) { return elementAt(index).deltaTime(); }
	public void setDeltaTimeAt(int deltaTime, int index) {
		elementAt(index).setDeltaTime(deltaTime); }

	public int length() {											// returns highest (deltaTime + length)
		int length = 0;
		for(int i = 0; i < size(); i++)
			if((deltaTimeAt(i) + lengthAt(i)) > length) length = deltaTimeAt(i) + lengthAt(i);
		return length; }

	public void setLength(int length) {				// multiplies all deltaTimes with a factor
		if(!lengthable) return;									// no effect if not lengthable
		double factor = length / (double)length();
		for(int i = 0; i < size(); i++) {
			setDeltaTimeAt((int)Math.round(deltaTimeAt(i) * factor), i);
			setLengthAt((int)Math.round(lengthAt(i) * factor), i);
		}}

	public void turnRound() {
		if(!lengthable) return;									// no effect if not lengthable
		int length = length();
		for(int i = 0; i < size(); i++) {
			if(elementAt(i) instanceof Lengthable) ((Lengthable)elementAt(i)).turnRound();
			if(!linked) setDeltaTimeAt(length - (deltaTimeAt(i) + lengthAt(i)), i); }
		if(!linked) super.turnRound(); }				// order in vector turned round


//****************************** MIDIContainer routines **********************************


	public int type() {
		if(lengthable) if(linked) return LINKED; else return FREE;
		else return ABSOLUTE; }
	public void setType(int type) {
		switch(type) {
			case LINKED: linked = true;
			case FREE: lengthable = true; break;
			case ABSOLUTE: lengthable = false; break;
			default: System.out.println("ERROR: unknown MIDIContainer-type."); System.exit(-1);
		}}

	public int position() {											// current position (end of last add)
		if(size() == 0) return 0; else return deltaTimeAt(size()-1) + lengthAt(size()-1); }
	public void setPosition(int position) { add(new Rest(position)); }

	public void append(MIDIObject object) { appendAt(object, position()); }
	public void appendAt(MIDIObject object, int position) {
		MIDIObject o = object.copy(); o.setDeltaTime(position); add(o); }

	protected void add(MIDIObject object) {		// overwritten: channel is set if channelable
		if(channelable && (object instanceof Channelable))
			((Channelable)object).setChannel(channel);
		super.add(object); }

	public MIDIObject copy() { return new MIDIContainer(this); }
}
